/* Mover for logrotated logfiles
 *
 * Written by Matthias Hensler, 2001/07/15
 * Updated: 2001/07/18: strip path from filename when moving
 *          2001/07/19: delete moved file
 *          2003/01/10: added exclude option
 *          2004/06/06: added largefile defines
 *          2006/07/30: added option to preserve mtime
 *          2006/08/02: preserve filename on empty pattern
 *                      preserve owner and mode
 *          2007/11/10: added mode to open() when used with O_CREAT
 * 
 * WSPse 2001-06, http://www.wspse.de/
 *
 * Licenced under GPL version 2
 * 
 * Author: matthias@wspse.de
 *
 * Contributions/Ideas:
 * Simon Klempert: preserving of mtime
 * Lubos Stanek: preserving of mode/owner, Make-/Specfile updates
 *
 *
 * Description: takes all files from command line and renames them to his old
 *              filename with a timestamp added. If the file has a numbered
 *              suffix (eg ".1", ".2", etc), this suffix is removed.
 *              If the file has a suffix like ".gz" or ".bz2", the suffix will
 *              be left untouched
 *
 *              if specified on command line, a file will be compressed by using
 *              gzip, bzip2 or similiar programs
 *
 *              if specified on command line, a file will be moved to an other
 *              directory
 *
 * Syntax: logmover [-v] [-f] [-n] [-p] [-m <dir>] [-bzip2] [-gzip] [-c <program>] [-e <name_1> ... <name_n> -xe] [-] file [file2] [file3] ...
 *
 * if "-v" is specified, the output will be verbose.
 * if "-f" is specified, the mover won't stop if a problem occurs.
 * if "-n" is specified, the program will delete files with are empty (0 bytes)
 * if "-p" is specified, the mtime, atime, mode and owner of the moved files will be preserved
 * if "-m <dir>" is specified, the files will be moved to <dir> later.
 * if "-bzip2" is specified, the files will be compressed with bzip2.
 * if "-gzip" is specified, the files will be compressed with gzip.
 * if "-c <program>" is specified, the files will be compressed with
 *    program <program>. <program> will get a filename as first argument,
 *    it should compress this filename, add a suffix, and delete the original
 *    file.
 * 
 * it's possible to use different compressmethods and different destdirs, eg:
 * logmover -m dir1 -bzip2 file1 file2 file3 -m dir2 -gzip file4 file5
 * if you have filenames like "-m" or so, you can specify "-" after the last option,
 * and no further options will be evaluated
 */

#define _LARGEFILE64_SOURCE 1
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <utime.h>

#define VERSION "0.07"
#define GZIP "gzip"
#define BZIP2 "bzip2"
#define TSDEF "YMD"
#define MAXBUF 4096

int prepare_pos(char *file, int *start, int *end)
{
  int i;

  i = strlen(file) - 1;

  while(1) {
    if(file[i] < '0' || file[i] > '9') {
      while(file[i] != '.' && i) {
	if(file[i] == '/') {
	  return 0;
	}
	i--;
      }
      if(i) i--;
    }
    if(! i) {
      return 0;
    }
    
    *end = i;
    
    while(file[i] != '.' && i>=0) {
      if(file[i] < '0' || file[i] > '9') {
	*end = -1;
      }
      i--;
    }
    
    if(i < 0) {
      return 0;
    }
    if(*end != -1) {
      *start = i;
      return 1;
    }
  }
}

char *prepare_name(char *old, time_t timestamp, char *tsformat)
{
  int start, end;
  char *new;
  char *timestr;
  char format[6];
  int i,len;
  struct tm *ltm;

  if(! old) {
    return NULL;
  }

  if(!tsformat || !*tsformat) return strdup(old);
    
  ltm = localtime(&timestamp);
  len = 0;
  for(i=0;i<strlen(tsformat);i++) {
    if(tsformat[i] == 'Y') {
      len += 4;
    } else if(tsformat[i] == 'y' || tsformat[i] == 'M' || tsformat[i] == 'D'
	      || tsformat[i] == 'h' || tsformat[i] == 'm' || tsformat[i] == 's') {
      len += 2;
    } else {
      len++;
    }
  }

  timestr = (char *) malloc(len + 10);
  if(! timestr) {
    perror("malloc");
    return NULL;
  }
  timestr[0] = 0;

  for(i=0;i<strlen(tsformat);i++) {
    switch(tsformat[i]) {
      case 'Y': sprintf(format, "%04d", ltm->tm_year+1900); break;
      case 'y': sprintf(format, "%02d", ltm->tm_year); break;
      case 'M': sprintf(format, "%02d", ltm->tm_mon+1); break;
      case 'D': sprintf(format, "%02d", ltm->tm_mday); break;
      case 'h': sprintf(format, "%02d", ltm->tm_hour); break;
      case 'm': sprintf(format, "%02d", ltm->tm_min); break;
      case 's': sprintf(format, "%02d", ltm->tm_sec); break;
      default: format[0] = tsformat[i]; format[1] = 0;
    }
    strcat(timestr, format);
  }
  
  new = (char *) malloc(strlen(old) + strlen(timestr) + 10);
  if(! new) {
    perror("malloc");
    free(timestr);
    return NULL;
  }
  
  if(! prepare_pos(old, &start, &end)) {
    sprintf(new, "%s.%s", old, timestr);
    free(timestr);
    return new;
  }

  i=0;
  while(i<start) {
    new[i] = old[i];
    i++;
  }
  new[i] = '.';
  new[i+1] = 0;
  strcat(new, timestr);
  strcat(new, old+end+1);

  free(timestr);
  return new;
}

int move_file (char *src, char *dest)
{
  int in, out;
  char buf[MAXBUF];
  int len;

  in = open(src, O_RDONLY);
  if(! in) {
    perror("open");
    fprintf(stderr, "%s\n", src);
    return 1;
  }

  out = open(dest, O_WRONLY | O_CREAT | O_TRUNC, 0644);
  if(! out) {
    perror("open");
    fprintf(stderr, "%s\n", dest);
    close(out);
    return 1;
  }

  while((len = read(in, &buf, MAXBUF))) {
    if(len == -1) {
      perror("read");
      fprintf(stderr, "%s\n", src);
      close(in);
      close(out);
      return 1;
    }
    if(write(out, &buf, len) != len) {
      perror("write");
      fprintf(stderr, "%s\n", dest);
      close(in);
      close(out);
      return 1;
    }
  }
  
  unlink(src);
  chmod(dest, 0644);
  return 0;
}

int logmove(char *file, char *destdir, char *compressor, char verbose, char delnull,
	    char overwrite, char *tsformat, char preserve)
{
  struct stat st;
  char filetyp;
  char *newfile, *newdir;
  char *command;
  char *pat;
  struct utimbuf utbuf;
  uid_t owner;
  gid_t group;
  mode_t mode;

  if(stat(file, &st)) {
    perror("stat");
    fprintf(stderr, "%s\n", file);
    return 1;
  }
  
  utbuf.modtime = st.st_mtime;
  utbuf.actime  = st.st_atime;
  owner         = st.st_uid;
  group         = st.st_gid;
  mode          = st.st_mode;
  filetyp   = 0;
  if(! S_ISREG(st.st_mode)) {
    if(! S_ISDIR(st.st_mode)) {
      fprintf(stderr, "\"%s\" is no regular file.\n", file);
      return 1;
    }
    filetyp = 1;
  }

  if(S_ISREG(st.st_mode) && (! st.st_size) && delnull) {
    if(unlink(file)) {
      perror("unlink");
      fprintf(stderr, "%s\n", file);
      return 1;
    }
    if(verbose) {
      printf("deleted \"%s\" because it was empty.\n", file);
    }
    return 0;
  }

  newfile = prepare_name(file, utbuf.modtime, tsformat);
  if(! newfile) {
    return 1;
  }
 
  if(tsformat && *tsformat) {
    if(! stat(newfile, &st)) {
      if(overwrite && verbose) {
	printf("overwriting \"%s\"\n", newfile);
      } else if(! overwrite) {
	printf("\"%s\" already exist\n", newfile);
	free(newfile);
	return 1;
      }
    }

    if(rename(file, newfile)) {
      perror("rename");
      fprintf(stderr, "%s -> %s\n", file, newfile);
      free(newfile);
      return 1;
    }
  }

  if(destdir) {
    if(filetyp) {
      fprintf(stderr, "moving directories is not supported in this version.\n");
      free(newfile);
      return 1;
    }
    newdir = (char *) malloc(strlen(destdir) + strlen(newfile) + 10);
    if(! newdir) {
      perror("malloc");
      free(newfile);
      return 1;
    }
    pat = strrchr(newfile, '/');
    if(pat) {
      pat++;
    } else {
      pat = newfile;
    }
    sprintf(newdir, "%s/%s", destdir, pat);

    if(! stat(newdir, &st)) {
      if(overwrite && verbose) {
	printf("overwriting \"%s\"\n", newdir);
      } else if(! overwrite) {
	printf("\"%s\" already exist\n", newdir);
	free(newfile);
	free(newdir);
	return 1;
      }
    }
    
    if(move_file(newfile, newdir)) {
      fprintf(stderr, "move failed: %s -> %s\n", newfile, newdir);
      free(newfile);
      free(newdir);
      return 1;
    }
    
    free(newfile);
    newfile = newdir;
  }

  if(preserve) {
    if(utime(newfile, &utbuf)) {
      fprintf(stderr, "preserving time failed for \"%s\"\n", newfile);
      perror("utime");
      return 1;
    }
    if(chown(newfile, owner, group)) {
      fprintf(stderr, "preserving owner failed for \"%s\"\n", newfile);
      perror("chown");
      return 1;
    }
    if(chmod(newfile, mode)) {
      fprintf(stderr, "preserving mode failed for \"%s\"\n", newfile);
      perror("chmod");
      return 1;
    }
  }

  if(compressor) {
    command = (char *) malloc(strlen(compressor) + strlen(newfile) + 10);
    if(! command) {
      perror("malloc");
      free(newfile);
      return 1;
    }
    sprintf(command, "%s \"%s\"", compressor, newfile);
    if(system(command)) {
      perror("system");
      fprintf(stderr, "%s\n", command);
      free(command);
      free(newfile);
      return 1;
    }
    free(command);
  }

  if(verbose) {
    printf("moved");
    if(compressor) {
      printf(" and compressed");
    }
    printf(" \"%s\" to \"%s\"\n", file, newfile);
    if(compressor) {
      printf(" (suffix and/or filename may be changed by compressor)\n");
    }
  }
   
  return 0;
}

int main(int argc, char **argv)
{
  int i,j;
  char *destdir;
  char *compressor;
  char *tsformat;
  char verbose, force, status, delnull, overwrite, evaluate, preserve;
  char gzip[] = GZIP;
  char bzip2[] = BZIP2;
  char tsdef[] = TSDEF;
  int exbegin, exend;

  destdir    = NULL;
  compressor = NULL;
  tsformat   = tsdef;
  verbose    = 0;
  force      = 0;
  status     = 0;
  delnull    = 0;
  overwrite  = 0;
  evaluate   = 1;
  preserve   = 0;
  exbegin    = -1;
  exend      = -1;

  if(argc == 1) {
    fprintf(stderr, "usage: %s [-v] [-f] [-n] [-o] [-p] [-m <destdir>] [-t <format>] [-bzip2]\n"
	            "      [-gzip] [-c <program>]\n       [-e <name_1> ... <name_n> -xe] [-] file\n"
		    "      [files...]\n detailed help with \"%s -h\"\n",
	   argv[0], argv[0]);
    exit(EXIT_FAILURE);
  }

  for(i=1;i<argc;i++) {
    if((!strcmp(argv[i], "-m")) && evaluate) {
      if(i+1 == argc) {
	fprintf(stderr, "%s: please specify a destination directory after \"-m\"\n", argv[0]);
	exit(EXIT_FAILURE);
      }
      destdir = argv[i+1];
      i++;
    } else if((!strcmp(argv[i], "-xm")) && evaluate) {
      destdir = NULL;
    } else if((!strcmp(argv[i], "-v")) && evaluate) {
      verbose = 1;
    } else if((!strcmp(argv[i], "-xf")) && evaluate) {
      verbose = 0;
    } else if((!strcmp(argv[i], "-f")) && evaluate) {
      force = 1;
    } else if((!strcmp(argv[i], "-xf")) && evaluate) {
      force = 0;
    } else if((!strcmp(argv[i], "-n")) && evaluate) {
      delnull = 1;
    } else if((!strcmp(argv[i], "-xn")) && evaluate) {
      delnull = 0;
    } else if((!strcmp(argv[i], "-o")) && evaluate) {
      overwrite = 1;
    } else if((!strcmp(argv[i], "-xo")) && evaluate) {
      overwrite = 0;
    } else if((!strcmp(argv[i], "-p")) && evaluate) {
      preserve = 1;
    } else if((!strcmp(argv[i], "-xp")) && evaluate) {
      preserve = 0;
    } else if((!strcmp(argv[i], "-e")) && evaluate) {
      if(i+1 == argc) {
	fprintf(stderr, "%s: please specify filenames to exclude after \"-e\"\n"
		"end your list by specifing \"-xe\"\n", argv[0]);
	exit(EXIT_FAILURE);
      }
      exbegin = i+1;
      i++;
      while(strcmp(argv[i], "-xe")) {
	if(i+1 == argc) {
	  fprintf(stderr, "%s: please end your exclude-list by specifing \"-xe\"\n", argv[0]);
	  exit(EXIT_FAILURE);
	}
	i++;
      }
      exend = i;
      if(exbegin == exend) exbegin = -1;
    } else if((!strcmp(argv[i], "-xe")) && evaluate) {
      exbegin = -1;
    } else if((!strcmp(argv[i], "-t")) && evaluate) {
      if(i+1 == argc) {
	fprintf(stderr, "%s: please specify a timestamp format after \"-t\"\n"
		"with: \"Y\" year 4-digits, \"y\" year 2-digits, \"M\" month\n"
		"      \"D\" day, \"h\" hours, \"m\" minutes, \"s\" seconds\n"
		,argv[0]);
	exit(EXIT_FAILURE);
      }
      tsformat = argv[i+1];
      i++;
    } else if(((!strcmp(argv[i], "-h")) || (!strcmp(argv[i], "-help"))
	       || (!strcmp(argv[i], "--h")) || (!strcmp(argv[i], "--help")))
	      && evaluate) {
      printf("Logmover " VERSION " by Matthias Hensler, http://www.wspse.de/\n"
	     "\n"
	     "renames logfiles, compresses them (if specified) and moves them to\n"
	     "another directory (if specified)\n"
	     "\n"
	     "usage: %s [options] [files] [options] [-] [files]\n"
	     "\n"
	     "options are the following:\n"
	     " -v       for verbose output\n"
	     " -f       to force execution of logmover, when moving\n"
	     "          of one file fails\n"
	     " -xf      to disable forcemode\n"
	     " -n       to delete files with 0 byte size\n"
	     " -xn      to disable deletion of nullbyte files\n"
	     " -o       to overwrite existing files\n"
	     " -xo      to disable overwriting existing files\n"
	     " -p       to preserve mtime, atime, mode and owner on moving\n"
	     " -xp      to disable the preserving of fileattributes\n"
	     " -m <dir> to move files to directory <dir>\n"
	     " -xm      to disable moving files to directory\n"
	     " -c <prg> to compress files with <prg>\n"
	     " -xc      to disable compression\n"
	     " -e <...> to specify a list of files to exclude, end the list with \"-xe\"\n"
	     " -xe      end an exclude-list or disable exclude-list\n"
	     " -gzip    shortcut for \"-c gzip\"\n"
	     " -bzip2   shortcut for \"-c bzip2\"\n"
	     " -t <ts>  use different timestamp (default \"YMD\")\n"
	     "          use \"%s -t\" to get a complete list\n"
	     "\n"
	     "a file will be renamed, moved and compressed with the options\n"
	     "specified before, so the options can be changed for a list of\n"
	     "files.\n"
	     "\n"
	     "options will be evaluated, until a single \"-\" marks the\n"
	     "end of options\n"
	     "\n\n\n"
	     "example:\n"
	     "%s -f -o -n -bzip2 /var/log/* -xo -xc -m /root/data /var/log/httpd/* \\\n"
	     "  -c gzip -xm -xf -t \"YDM-h:m\" - /var/log/ncftpd/*\n"
	     "will rename all files in /var/log, compresses them with bzip2, deletes\n"
	     "files with 0 byte size, overwrite existing files and don't stop if an\n"
	     "error occurs.\n"
	     "then it will rename all files in /var/log/httpd, moves them to /root/data,\n"
	     "with disabled compression, and does not overwrite existing files.\n"
	     "at last it will rename all files in /var/log/ncftpd with the timestamp\n"
	     "\"YDM-h:m\", compresses them with gzip, does not move them to /root/data\n"
	     "and stops if an problem occurs, furthermore it doesn't accept any options,\n"
	     "so a file named \"-v\" in the directory won't make any problem.\n\n"
	     "example for exclude-lists:\n"
	     "%s -e /var/log/maillog /var/log/messages -xe /var/log/* \\\n"
	     "  -xe /var/log/httpd/*\n"
	     "will rename all files in /var/log, without touching maillog and messages. Then\n"
	     "disable the exclude-list and rename the files in /var/log/httpd\n"
	     , argv[0], argv[0], argv[0], argv[0]);
      exit(EXIT_SUCCESS);
    } else if((!strcmp(argv[i], "-bzip2")) && evaluate) {
      compressor = bzip2;
    } else if((!strcmp(argv[i], "-gzip")) && evaluate) {
      compressor = gzip;
    } else if((!strcmp(argv[i], "-c")) && evaluate) {
      if(i+1 == argc) {
	fprintf(stderr, "%s: please specify a compressprogram after \"-c\"\n", argv[0]);
	exit(EXIT_FAILURE);
      }
      compressor = argv[i+1];
      i++;
    } else if((!strcmp(argv[i], "-xc")) && evaluate) {
      compressor = NULL;
    } else if((!strcmp(argv[i], "-")) && evaluate) {
      evaluate = 0;
    } else {
      j = exbegin;
      while(j > 1 && j <= exend) {
	if(!strcmp(argv[i], argv[j])) {
	  if(verbose) {
	    fprintf(stderr, "%s: skipped \"%s\" because file is in exclude-list\n", argv[0], argv[i]);
	  }
	  j = -2;
	  break;
	}
	j++;
      }
      if(j != -2) {
	if(logmove(argv[i], destdir, compressor, verbose, delnull, overwrite, tsformat, preserve)) {
	  if(! force) {
	    exit(EXIT_FAILURE);
	  }
	  status = 1;
	}
      }
    }
  }

  if(status) {
    return EXIT_FAILURE;
  } else {
    return EXIT_SUCCESS;
  }
}

/* WSPse 2001-06 - http://www.wspse.de/ - matthias@wspse.de */
